# RDSR
ENHANCED_STRUCTURED_REPORT_UID = "1.2.840.10008.5.1.4.1.1.88.22"
XRAY_RADIATION_DOSE_STRUCTURED_REPORT_UID = "1.2.840.10008.5.1.4.1.1.88.67"
RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS = [
    ENHANCED_STRUCTURED_REPORT_UID,
    XRAY_RADIATION_DOSE_STRUCTURED_REPORT_UID,
]

# Secondary Capture
SECONDARY_CAPTURE_UID = "1.2.840.10008.5.1.4.1.1.7"
SECONDARY_CAPTURE_SOP_CLASS_UIDS = [SECONDARY_CAPTURE_UID]



from typing import List, Optional, Tuple, Union

from numpy import (
    abs,
    absolute,
    add,
    arctan,
    array,
    ceil,
    clip,
    divide,
    exp,
    median,
    multiply,
    ones,
    pi,
    power,
    sort,
    sqrt,
    subtract,
    sum,
    vectorize,
    zeros,
)
from scipy import linalg
from scipy.optimize import curve_fit

from .point import IntPoint, Point

LINE_COORDINATE = Union[int, float]
FIT_DATA = Union


class Line:
    def __init__(self, x0: LINE_COORDINATE, y0: LINE_COORDINATE, x1: LINE_COORDINATE, y1: LINE_COORDINATE):
        if not all([(isinstance(obj, int) or isinstance(obj, float)) for obj in [x0, y0, x1, y1]]):
            raise TypeError("All line coordinates must be either integers or floats")

        self.slope = (y1 - y0) / (x1 - x0)
        self.constant = y0 - self.slope * x0
        self.angle = arctan(self.slope) * 180 / pi


def get_shortest_line_point_dist(line: Line, point: Union[Point, IntPoint]):
    if not isinstance(line, Line):
        raise TypeError("line must be a Line object")

    if not isinstance(point, Point) and not isinstance(point, IntPoint):
        raise TypeError("point must be either a Point or IntPoint instance")

    distance = absolute(line.slope * point.x + line.constant - point.y) / sqrt(line.slope ** 2 + (-1) ** 2)
    if line.angle <= 45 or line.angle >= 135:
        if point.y < (line.slope * point.x + line.constant):
            distance = -distance
    else:
        if point.x < (point.y - line.constant) / line.slope:
            distance = -distance

    return distance


def third_deg_polynomial_fit_func(x, a, b, c, d):
    return add(multiply(a, power(x, 3)), add(multiply(b, power(x, 2)), add(multiply(c, x), d)))


def get_third_degree_polynomial_fit(x: List[Union[float, int]], y: List[Union[float, int]]):
    """Fits data to a 3rd degree polynomial on the format y = a*x^3 + b*x^2 + c*x + d. Returns y-values calculated with
    the optimized function and the z, b, c, and d parameters in a dict.

    Args:
        x: Values on the x-axis
        y: Noisy y-values to fit to the function

    Returns:
        A dictionary on with the keys a, b, c, d and y, where y is the optimized y values.

    """
    popt, pcov = curve_fit(f=third_deg_polynomial_fit_func, xdata=x, ydata=y)
    optimized_y = third_deg_polynomial_fit_func(x, *popt)

    return dict(a=popt[0], b=popt[1], c=popt[2], d=popt[3], y=optimized_y)


def _bell_kernel_function(x: Union[int, float], x0: Union[int, float], tau: Optional[float] = 0.005):
    return exp(-divide(power(subtract(x, x0), 2), multiply(2, tau)))


def _test_function(x, x1, ind):
    output = sort(abs(subtract(x, x1)))[ind]
    print(output)
    return output


def lowess(
    x: List[Union[float, int]], y: List[Union[float, int]], f: Optional[float] = 2.0 / 3.0, iter: Optional[int] = 3
):
    """lowess(x, y, f=2./3., iter=3) -> yest
    Lowess smoother: Robust locally weighted regression.
    The lowess function fits a nonparametric regression curve to a scatterplot.
    The arrays x and y contain an equal number of elements; each pair
    (x[i], y[i]) defines a data point in the scatterplot. The function returns
    the estimated (smooth) values of y.
    The smoothing span is given by f. A larger value for f will result in a
    smoother curve. The number of robustifying iterations is given by iter. The
    function will run faster with a smaller number of iterations.
    """
    n = len(x)
    r = int(ceil(f * n))
    tf = vectorize
    h = [sort(abs(subtract(x, x[i])))[r] for i in range(n)]
    w = clip(abs(divide((subtract(x[:, None], x[None, :])), h)), 0.0, 1.0)
    w = power(power(subtract(1, w), 3), 3)
    yest = zeros(n)
    delta = ones(n)
    for iteration in range(iter):
        for i in range(n):
            weights = multiply(delta, w[:, i])
            b = array([sum(multiply(weights, y)), sum(multiply(weights, multiply(y, x)))])
            A = array(
                [[sum(weights), sum(multiply(weights, x))], [sum(weights * x), sum(multiply(multiply(weights, x), x))]]
            )
            beta = linalg.solve(A, b)
            yest[i] = add(beta[0], multiply(beta[1], x[i]))

        residuals = subtract(y, yest)
        s = median(abs(residuals))
        delta = clip(residuals / multiply(6.0, s), -1, 1)
        delta = power(subtract(1, power(delta, 2)), 2)

    return yest



from pathlib import Path
from typing import List, Optional, Union

import pydicom
from pydicom import FileDataset

from .dose_matrix import DoseMatrix
from ..constants.SopClassUids import (
    RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS,
    SECONDARY_CAPTURE_SOP_CLASS_UIDS,
)
from .ct import CtSeries
from .dicom_series import DicomSeries
from .dose_report_class import DoseReport
from .projection import ProjectionSeries
from ..helpers.check_path_is_valid import check_path_is_valid_path


class DicomStudy:
    """A class to manage DICOM files connected by a Study Instance UID

    Args:
        study_instance_uid : The study instance UID of the DICOM study object that is to be created

    Attributes:
        StudyInstanceUid: The study instance UID of the DICOM study object
        Series: List of DicomSeries objects containing detailed information abouts series included in the study
        Manufacturer: The manufacturer of the machine used in acquiring the image/-s
        ManufacturerModelName: The name of the machine, as given by the manufacturer, used in acquiring the image/-s

    Raises:
        TypeError: if study_instance_uid is not a string

    """

    def __init__(self, study_instance_uid: str):
        if not isinstance(study_instance_uid, str):
            raise TypeError("study_instance_uid must be a string")
        # Metadata
        self.StudyInstanceUid: str = study_instance_uid
        self.Series: List[Union[DicomSeries, CtSeries, ProjectionSeries]] = []
        self.Manufacturer: Optional[str] = None
        self.ManufacturerModelName: Optional[str] = None
        self.DoseReports: Optional[DoseReport] = DoseReport()

    def add_file(self, file: Union[Path, str], dcm: Optional[FileDataset] = None) -> None:
        """Add the DICOM file to the DicomStudy object after validating the study instance UID

        Args:
            file: Path to where the file to be added is stored on disc
            dcm: The DICOM-file imported to a FileDataset object

        Raises:
            InvalidDicomError: If the given file is not a valid DICOM file
            TypeError: If the file is not a valid path
            ValueError: If the file does not have the same study instance UID as the StudyInstanceUID of the object

        """
        file = check_path_is_valid_path(path_to_check=file)

        if dcm is None:
            dcm = pydicom.dcmread(fp=str(file.absolute()), stop_before_pixels=True)

        if dcm.StudyInstanceUID != self.StudyInstanceUid:
            raise ValueError(f"The given DICOM file is not part of the study {self.StudyInstanceUid}")

        self.Manufacturer = dcm.Manufacturer
        self.ManufacturerModelName = dcm.ManufacturerModelName

        if dcm.SOPClassUID in RADIATION_DOSE_STRUCTURED_REPORT_SOP_CLASS_UIDS + SECONDARY_CAPTURE_SOP_CLASS_UIDS:
            self.DoseReports.add_file(dataset=dcm)
            return

        try:
            index = [obj.SeriesInstanceUid for obj in self.Series].index(dcm.SeriesInstanceUID)
        except ValueError:
            if dcm.Modality == "CT":
                self.Series.append(CtSeries(series_instance_uid=dcm.SeriesInstanceUID))
            elif dcm.Modality == "RTDOSE":
                self.Series.append(DoseMatrix(file=file, dcm=dcm))
            else:
                self.Series.append(ProjectionSeries(file=file, dcm=dcm))
            index = [obj.SeriesInstanceUid for obj in self.Series].index(dcm.SeriesInstanceUID)

        self.Series[index].add_file(file=file, dcm=dcm)



import logging
from typing import Dict, List, Optional, Union

import numpy as np
from scipy.stats import sem

from ..helpers.point import Point
from ..helpers.voxel_data import VoxelData
from ..roi.square_roi import SquareRoi

FloatListType = Union[float, List[float]]

LCD_CONSTANT = 3.29  # Value from A. Radice et al. (2016) Physica Medica (32)

log = logging.getLogger(__name__)


def lcd_statistical(stderr: FloatListType) -> FloatListType:
    """Implements a statistical method for determining the low contrast detectability (LCD) as mentioned in
    "A STATISTICAL METHOD FOR LOW-CONTRAST DETECTABILITY ANALYSIS IN ANGIOGRAPHY SYSTEMS. A. Radice, N. Paruccini,
    C. Spadavecchia, R. Villa, A. Baglivi and A. Crespi (2016) Physica Medica (32).
    https://doi.org/10.1016/j.ejmp.2016.07.728" This specifies that the low contrast detectability as 3.29 x sigma,
    where sigma is the standard error of the mean of 100 square ROIS of a side corresponding to the "hole" size that the
    LCD is determined for.

    Args:
        stderr: Standard error of 100 ROIs given as a single value or a list of the standard error of each ROI

    Returns:
        Low contrast resolution as a float (if stderr is a float) or a list of floats (if stderr is a list of floats)
    """
    if not isinstance(stderr, float) and not isinstance(stderr, list):
        raise TypeError("stderr must be given as a float or a list of floats")

    if isinstance(stderr, float):
        return LCD_CONSTANT * stderr

    return [LCD_CONSTANT * val for val in stderr]


def lcd_statistical_random(
    analysis_matrix: np.ndarray,
    pixel_size: VoxelData,
    object_size: float,
    rois: Optional[int] = 1000,
    return_rois: Optional[bool] = False,
) -> Dict[str, Union[str, float, List[float]]]:
    """Implements a statistical method for determining the low contrast detectability (LCD) or low contrast limit (LCL)

    Args:
        analysis_matrix: Matrix of the homogeneous image area that is to be analysed
        pixel_size: The voxel data for the analysis_matrix
        object_size: The side of the object to analyse for in mm
        rois: The number of ROIs to include in the analysis

    Returns:
        Dictionary containing the mean, standard deviation, standard error and LCD on the form
        {
          'Hole Diameter': str,                  # (mm)
          'ROI Box Size': str,                   # (pixels)
          'Mean': float,                         # Mean value
          'Std Error Mean': float,               # Standard error of the mean
          'SD': float,                           # Standard deviation
          'Error SD': float,                     # Standard error of the standard deviation
          'Perc contrast At 95 Perc CL': float,  # Percent contrast at 95% contrast level
          'ROIs': Optional[List[SquareRoi]],     # A list of all ROIs used or None
        }

    """
    if not isinstance(analysis_matrix, np.ndarray):
        raise TypeError("The analysis matrix must be a numpy ndarray")

    if not isinstance(pixel_size, VoxelData):
        raise TypeError("The pixel_size must be given as a VoxelData object")

    if not isinstance(object_size, float):
        raise TypeError("Object size must be a float")

    if not isinstance(rois, int):
        raise TypeError("The number of ROIs (rois) must be an integer")

    output = {
        "Hole Diameter": f"{object_size:.2f}mm",
        "ROI Box Size": f"{int(np.floor(object_size / pixel_size.x))}pixel",
        "Mean": float(),
        "Std Error Mean": float(),
        "SD": float(),
        "Error SD": float(),
        "Perc Contrast At 95 Perc CL": float(),
    }

    result = {
        "Mean": [],
        "SD": [],
        "Std Error Mean": [],
        "ROIs": [],
    }

    log.info(f"Calculating LCD from a set of {rois} ROIs of size {object_size:.3f} mm")

    # Place ROIs and get ROI statistics
    roi_size = [np.floor(np.divide(object_size, pixel_size.x)), np.floor(np.divide(object_size, pixel_size.y))]
    structured_rois = [analysis_matrix.shape[1] // roi_size[0], analysis_matrix.shape[0] // roi_size[1]]
    random_rois = int(np.floor(rois - structured_rois[0] * structured_rois[1]))

    # Determine the mean values and standard deviation for a structured ROI set
    for col in range(np.int(structured_rois[0])):
        for row in range(np.int(structured_rois[1])):
            roi = SquareRoi(
                center=Point(
                    x=int(col * roi_size[0] + np.ceil(roi_size[0] / 2)),
                    y=int(row * roi_size[1] + np.ceil(roi_size[1] / 2)),
                ),
                height=object_size,
                width=object_size,
                pixel_size=pixel_size,
                resize_too_big_roi=True,
            )
            result["Mean"].append(roi.get_mean(analysis_matrix))
            result["SD"].append(roi.get_stdev(analysis_matrix))
            result["Std Error Mean"].append(roi.get_std_error_of_the_mean(analysis_matrix))
            result["ROIs"].append(roi)

    if random_rois > 0:
        # Determine the mean value and standard deviation in a set of randomly placed ROIs
        for _ in range(random_rois):
            valid_roi = False
            n = 0
            while not valid_roi and n < 1000:
                roi = SquareRoi(
                    center=Point(
                        x=int(np.random.randint(0, analysis_matrix.shape[1]) + np.ceil(np.divide(roi_size[0], 2))),
                        y=int(np.random.randint(0, analysis_matrix.shape[0]) + np.ceil(np.divide(roi_size[1], 2))),
                    ),
                    height=object_size,
                    width=object_size,
                    pixel_size=pixel_size,
                    resize_too_big_roi=True,
                )
                tmp_mean = roi.get_mean(analysis_matrix)
                tmp_sd = roi.get_stdev(analysis_matrix)
                tmp_sem = roi.get_std_error_of_the_mean(analysis_matrix)

                n += 1
                valid_roi = not np.isnan(tmp_mean) and not np.isnan(tmp_sd) and not np.isnan(tmp_sem)

            if valid_roi:
                result["Mean"].append(tmp_mean)
                result["SD"].append(tmp_sd)
                result["Std Error Mean"].append(tmp_sem)
                result["ROIs"].append(roi)

    # Calculate the mean value and related values from the ROIs
    output["Mean"] = np.mean(result["Mean"])
    output["Std Error Mean"] = np.mean(result["Std Error Mean"])
    output["SD"] = np.mean(result["SD"])
    output["Error SD"] = sem(result["SD"])
    output["Perc Contrast At 95 Perc CL"] = lcd_statistical(stderr=output["Std Error Mean"])

    if return_rois:
        output["ROIs"] = result["ROIs"]

    return output
